
;**********************************************************************
;                                                                     *
;    Files Required: P16F886.INC                                      *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Notes:        CountClickL & simple Debouncer
;Latch is 1 when Sw is closed.
;Latch is 0 when Sw is open.
;ClickL is the number of switch presses. counts to 255.       
; Contains a binary to decimal routine for decimal digital display
; Displays via a Shift Register eg (74HC595 or TPIC6C595) with up to  4 i/o driving the 4 digit CA lines  and strobing the display as well as the switches.                                      *
; Column select pins also act as comms fror Shift Register and strobe the tactile switches - triple duty
;**********************************************************************


	list		p=16f886		; list directive to define processor
	#include	"P16F886.inc"		; processor specific variable definitions
	
	
		__CONFIG    _CONFIG1,  _IESO_OFF & _FCMEN_OFF & _PWRTE_ON  & _WDT_OFF & _INTOSCIO & _BOR_ON & _LVP_OFF & _CP_OFF &_CPD_OFF & _MCLRE_OFF &_INTRC_OSC_NOCLKOUT; ; 0x23E4 binary equivalent.
	__CONFIG    _CONFIG2,	_WRT_OFF & _BOR40V 
;***** VARIABLE DEFINITIONS

Delay1		EQU  	0x20		
Delay2		EQU 	0x21		
Switch		EQU 	0x22		
Latch		EQU 	0x23	
ClickL 		EQU 	0x24
ClickH		EQU	0x25
Units		EQU 	0x26	
Tens		EQU 	0x27	
Hundreds	EQU 	0x28
Thousands	EQU	0x29	
TempL		EQU 	0x2a
TempH		EQU	0x2b
FixedL		EQU	0x2c
FixedH		EQU	0x2d
Tmp		Equ	0x2e
Tmp1		Equ	0x2f
Offset		Equ	0x30
Digitstrobe	Equ	0x31 ; column select (7 segment)
Sw_state	Equ	0x32 ; switch states.

; Setup which pins assigned to which (column select) Digits here. No restriction other than all pins are Dig. outs.
#define 	Digitclock	PORTC,7
#define 	Unitsdriver	PORTC,6 ; column 0
#define 	Tensdriver	PORTC,5;column 1
#define 	Hundredsdriver	PORTC,6; make this PIN same as Unitsdriver in a 2 Digit display to reduce flicker. column 2
#define	Thousandsdriver	PORTC,5; make this PIN same as Tensdriver in a 2 Digit display. column 3

CONSTANT DIGITS = 2; set this to 2 or 4 depending on the # of digits in the display.

;**********************************************************************
	Org	0x000				; processor reset vector
 	goto	Main			; go to beginning of program
 	

Main
	
;Input PORTA setup (RA3)
	banksel	ANSEL
	clrf ANSEL; make all PORTA digital
	clrf ANSELH; PORTB all digital
	Banksel TRISA
	clrf	TRISA ;make all of port A outputs
	BSF TRISE,3 ; Set RE3, pin as an input. MCLR off.


;PORTC  setup
	;BankSel TRISC  same bank as TRISA
	clrf TRISB
	CLRF TRISC	; make portC outputs
	BankSel ClickL ; back to general variable bank
	movlw 0x3f	; '0' shape for 7 segment display.
	movwf PORTC; init PORTC to 7seg '0'
	movwf Units ; 7 seg '0'
	movwf Tens ; 7 seg '0'
	movwf Hundreds ; 7 seg '0'
	movwf Thousands ;""
		
	
;init variables
	CLRF Switch
	INCF Switch,F
	CLRF Latch
	clrf ClickH
	Movlw .0 ; ClickL sould start at zero, use this value for testing.
	movwf ClickL 

; Main operating loop
Longloop
	CLRF	Delay1
	CLRF 	Delay2 
				
LOOP	
	
	DECFSZ Delay1,F ; 1 inst cycle, when  delay1=0 => test = true skip next line.
	GOTO LOOP		; 2 inst cycles, this  is looped 256 times before skipping to next line => 768 microsec at 1 MIP
	CALL GETSWITCH ; this is the core loop, use to monitor switch.
	CALL Strobe ; strobe the display based on 2 lsb's of Delay2
	DECFSZ	Delay2,F ; 1 inst cycle
	GOTO LOOP		; 2 inst cyles, by adding the loops = 256 * 768 + (256*3) = 257* 768 = 197376 microsec = .2 sec.  Simple Substitute for  ISR time control.
	Goto Longloop ; CONTINUE

	
;***************************
Strobe	 ; this routine strobes the 2 or 4 digit display to match  digit & data shown. Also strobes the tactile (N.O.) switches.
	;Tensdriver & Unitsdriver are reused for the Shiftregister serial xfer.
	; SEARCH & UPDATE all  'D_invert' commented code lines  (13 of them) should the common (cathode or anode) drive line logic need to be inverted based the type of display used.
	; since each LED segment consumes the rated LED power (eg. 20mA continuous or 50mA pulsed in <1mS), Max common mode drain (incl. decimal) = 8x rated, eg. 160mA
	; This loading requires a common mode driver transistor with resultant driver line logic inversion.
	; each LED segment dropping resistor for a 10mA, 2 Vf RED display @5V supply=> 5-2=3V, V=IR => R=V/I => 3/.01 = 300 ohms.
	; @ 1/2 duty cycle overdriving with a 270ohm is reasonable, since all segments will rarely be lit.
			
	incf Digitstrobe,f ; increment the display digit we are updating.
	;Next, make all common drive lines 'disabled', hi for CC or Low for CA. The assumed driver transistor on the common lines inverts this logic! This permits setup of each digit's pattern data without display flicker.
	;________________________________________________________
	bcf Thousandsdriver ;  *** D_invert
	bcf Hundredsdriver; *** D_invert
	bcf Tensdriver;  *** D_invert
	bcf Unitsdriver;  *** D_invert
	;________________________________________________________
	clrf Tmp1 ; clear bitmask byte - local temp variable , used to sequentially track which digit to strobe.
	movf Digitstrobe,w ; load counter - informs which digit to strobe.
	andlw b'00000011' ; checking for bits 1 & 0 only for counting  (up to 4) digits 0,1,2,3.  Modify this to '00000111'  and add sequential GPR display buffers if u need to go to 8 digits.
	movwf Offset; local digit select counter.
	incf Offset,f; correct for next decf cmd.
;'bbbbXXXX' bitmask will be used track digit drive pins. (Buffer 4,3,2,1 : thou,hund,tens,units)
	movlw Thousands ; highest digit value, display buffer. GPR for the digits must be in a continuous ascending order sequence.
	movwf FSR ; setup indirect addressing for Display Buffers.
	
	;Next 8 lines configures for a 2 digit display or a 4 digit. The defines above determine  how this works
	; For a 2 digit display we simply repeat the lower 2 digits so the 4 digit code flow remains intact.
	movlw DIGITS
	sublw 4
	skpnz; if 2 digits, else
	goto Chk_Dcount ; 
	 movf Tens,w
	 movwf Thousands
	 movf Units,w
	 movwf Hundreds
	
Chk_Dcount
	bsf Tmp1,4 ; initialize the bitmask to 0001XXXX
	bcf STATUS,C; clr C
	decfsz Offset,f
	goto ShiftDigit
	bsf Offset,3 ; make offset=8 for # of bits in Character pattern to send serially in Transmit routine called next.
	movf INDF,w ; get correct buffer. Always complement this pattern for CA. => comf INDF,w ; ****
	movwf Tmp ; store in Tmp, clearing this GPR here can cause an empty display, useful for PWM Dimming.


	call Transmit ; setup Shift register with  character data pattern.
	; Code below selects which common cathode drive line to strobe.
	;Next 4 lines configures for a 2 digit display or a 4 digit. The defines above determine  how this works
	; For a 2 digit display we simply repeat the lower 2 digits so the 4 digit code flow remains intact.
	movlw DIGITS
	sublw 4
	skpz
	goto Twodigits
		
	btfsc Tmp1,4 
	bsf Thousandsdriver ;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic) *** D_invert
	btfsc Tmp1,5
	bsf Hundredsdriver;  A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic) *** D_invert
Twodigits;
	btfsc Tmp1,6 
	bsf Tensdriver;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic) *** D_invert
	btfsc Tmp1,7
	bsf Unitsdriver;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic) *** D_invert
	skpnz
	goto PollSwitches ; replaces a 'return' cmd.
	btfsc Tmp1,4 
	bsf Tensdriver;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic) *** D_invert
	btfsc Tmp1,5
	bsf Unitsdriver;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic) *** D_invert
PollSwitches; Each  digit drive (column select) for the 7 segment can also drive a switch. Sw_state is setup next so that bits 0,1 match switches 0,1 status, etc. Sw_state can be used in a debounced parallel keypad routine.
	swapf Tmp1,w ; XXXXbbbb , bits 0-3 rolling column select for switches. Max # of switches = # of columns. A left sweep => bit 0 = Sw tied to units col, bit1 = sw tied to tens col etc.
	iorwf Sw_state,f; preset the  switch state to '1' = pressed, bit0 = sw0 , bit1 = sw1 etc.
	btfss PORTE,3 ; Make this 'btfsc'for a CA display. D_invert ***. Test switch bus, vdd pull up, goes to (CA) 0, or (1) CC  ,due to transistor drive inversion, when  sw is pressed. Switches tied to pins matching LED column select 'defines' above. A left sweep => bit 0 = Sw tied to Units column pin, bit1 = sw tied to Tens col etc.
	xorwf Sw_state,f; toggle the preset sw state to 'off' if sw. not pressed, leaving other bits intact. This 'strobes' the switches states.
	return
	
ShiftDigit; rotate bitmask to select digit to enable.
	rlf Tmp1,f
	decf FSR,f ; point to current display buffer
	btfss STATUS,C
	goto Chk_Dcount+2
	goto Chk_Dcount; reinit the mask.

Transmit; 
	rlf Tmp,f ; rotate MSB into carry bit, this routine transmits serial data from  MCU to high current display buffer.
	;setting serial data to the output serial data line
	BCF Tensdriver ;	s2p_serial_data_out
	BTFSC STATUS,C	;check the rotated out bit from the MSB
	BSF Tensdriver ;	s2p_serial_data_out


	BCF Unitsdriver ;	s2p_serial_clock ;initialise the serial clock
	BSF Unitsdriver ;	s2p_serial_clock ;clock the serial data in on positive edge


	decfsz Offset,f ; test Offset
	goto Transmit ; send another bit if Offset not yet zeroed, else

SERIAL2PARALLEL_END

	;Latch the new data onto the output buffer - Light the 7segment.
	BCF Digitclock ;s2p_Latch_clock	;initialise the Latch clock
	BSF Digitclock ;s2p_Latch_clock	;clock Latch the clocked serial data on positive edge
	; this also drives the attached Output Enable Low, suspending the display. Can be used to PWM dim.
	BcF Tensdriver ; disable Tens 7 segment, enabled upon return from sub.**;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic, D_invert) ***
	BcF Unitsdriver ; disable Units 7segment , enabled upon return from sub**;   A pnp (CA) or npn (CC) transistor common driver is assumed (inverts logic, D_invert) ***
	BCF Digitclock; make low to enable the  Display, as the OE is hooked to this line.

	;serial_clock or Latch_clock should be at logic 1 at any one time so that output will not be disturbed
	;the two clock line should not be at logic 0 at the same time or else the output will be affected by the data line.

	return
;*********************************************
GETSWITCH ;debounce routine of Switch 
	BTFSS Sw_state,0 ;  skip if HI (SW is CLOSED)
	CALL OPEN
	BTFSC Sw_state,0 ; , skip if lo (SW is OPEN)
	CALL CLOSED
	btfsc Sw_state,1
	clrf ClickL;reset count
	call Bindec
	RETURN
	
OPEN ; When Switch is Open
	DECFSZ Switch,F; Debounce Countdown if switch is open.
	Return ; return if switch value is non zero
	Incf Switch,f ; make switch minimum 1
	CLRF Latch ; clear onlatch as Switch has reached lowest count.
	RETURN
	
CLOSED: ; When switch is closed - Debounce delay
	INCF Switch,F ; Debounce count up if switch is on/closed
	MOVF Switch,w
	SUBLW .50 ; Set to 50 to get 50-w, @ 8 Mhz this delay is about 20ms.
	BTFSC STATUS,C ; DO NOT RETURN IF CARRY CLR
	GOTO RET ; RETURN on carry set (w<50) =>return as count not high enough.
	BSF Latch,1 ; set latch,1 to 1 if Switch =51
	CALL CLIK	
	DECF Switch,F ; keep Switch from rising past 51.
RET	RETURN
	
CLIK	; Track number of switch ClickL
	BTFSC Latch,0 ;skip next inst. if latch,0 is CLR. (if latch is 0)
	GOTO NOCOUNT
	INCF ClickL,f ; If Latch,0 is clear,then this is a state change from switch open => increment ClickL
	btfsc STATUS,Z ; test Z flag, skip if clear.
	incf ClickH,f ; execute if ClickL=0,else
	MOVLW .1
	MOVWF Delay2 ; SET Delay2 to .1 to force update of portC.
	CALL Bindec ; create a decimal number for the 7 seg display

NOCOUNT	BCF STATUS,C ; clear the carry
		RRF Latch,f ; make bit 1 into bit 0 to update Latch status for switch open/closed testing.
		RETURN

Bindec	;calculate decimal# for each 7seg led block
	clrf Thousands
	clrf Hundreds
	clrf Tens
	clrf Units
	movf ClickL,w
	movwf TempL
	movf ClickH,w
	movwf TempH
	call Thou ; get thousands in decimal
	movlw .10 ; check for 10,000 or more
	subwf Thousands,w;  if Thousands < 10, carry is clr.
	btfsc STATUS,C ;
	goto Overrange ;
	Call Hundred ; get hundreds
	call Ten ; get tens & units.
	return
Overrange
	movlw 0x79 ; E pattern
	movwf Thousands
	movlw b'01010000' ; r pattern
	movwf Hundreds
	Movwf Tens
	movlw 0x06 ; '1' pattern (error 1)
	movwf Units
	return
	
	

Thou	movlw .3 ; high byte set to 3 = 768	
	movwf FixedH ; high byte of # to be subtracted.
	movlw .232 ; low byte
	movwf FixedL ; 768 + 232 = 1000's column

ThouCount
	call Subtract
	xorlw .1 ; check for returned wreg value. A  zero occurs if a 1 was returned
	btfss STATUS,Z ; branch if 1 was returned, else
	goto Fixremainder ; due to last subtraction the remainder went -ve, so we fix this.
	incf Thousands,f
	goto ThouCount

Fixremainder ;restore remainder to continue calculating for next decimal column.
	movf    FixedL,w ; low byte
        addwf   TempL,f ; TempL=TempL + FixedL
        movf    FixedH,w ; hi byte
        btfsc   STATUS,C ; check for carry, set if  TempL + FixedL >255
        incfsz  FixedH,w ; Increment due to carry set ,if zero skip next.
        addwf   TempH,f	; TempH=TempH+FixedH (+1 if carry was set)
	return

Hundred	movlw 0 ; high byte set to 0	
	movwf FixedH ; high byte of # to be subtracted.
	movlw .100 ; low byte
	movwf FixedL ; 0 + 100 = 100's column

HdrdCount
	call Subtract
	xorlw .1 ; check for returned wreg value. A  zero occurs if a 1 was returned
	btfss STATUS,Z ; branch if 1 was returned, else
	goto Fixremainder ; due to last subtraction the remainder went -ve, so we fix this.
	incf Hundreds,f
	goto HdrdCount

  
Ten	movlw .10
	subwf TempL,f; test for 10 if carry clr wreg is under 10
	btfsc STATUS,C ;skip if carry clr (-ve result) , else
	incf Tens,f; incr the tens column
	btfsc STATUS,C ; skip if carry clr (-ve result), else
	goto Ten ; now check for 20,30 etc
	movf TempL,w
	addlw .10; as last calc was -ve recreate +ve number.
DispLEDS
	call Dpattern
	Movwf Units ; store 7 seg display pattern for units
	movf Tens,w	
	call Dpattern
	movwf Tens; store 7 seg display pattern for tens
	movf Hundreds,w
	call Dpattern
	movwf Hundreds; store 7 seg display pattern for tens
	movf Thousands,w
	call Dpattern
	movwf Thousands	
	return
Subtract ; 2byte subtraction to determine decimal 7 seg columns	
	movf    FixedL,W
        subwf   TempL,f ; TEMPL= TEMPL-FIXEDL
        movf    FixedH,W
        btfss   STATUS,C ; branch if TEMPL>FIXEDL, else
        incfsz  FixedH,W ; setup to reduced TempH by 1 extra (sim. a borrow).
        subwf   TempH,f  ;TempH=TempH-FixedH
	btfss	STATUS,C ;branch if TempH>FixedH, else
	retlw .0 ; if carry is clr return with a 0 in wreg
	retlw .1 ; if carry is set return with a 1 in wreg.


Dpattern	addwf PCL,f ; offset prog counter to lookup which # to be displayed
	retlw 0x3f ; code for 0, bit 7 is the dot. Common cathode. Invert values then clr bit 7 (dot) for common anode.
	retlw 0x06; 1
	retlw 0x5b; 2
	retlw 0x4f ;3
	retlw 0x66 ;4
	retlw 0x6d ;5
	retlw 0x7d ;6
	retlw 0x07 ;7
	retlw 0x7f ;8
	retlw 0x67 ;9
	;retlw 0x77 ;A
	;retlw 0x7c ;b
	;retlw 0x39 ;C
	;retlw 0x5e ;d
	;retlw 0x79 ;E
	;retlw 0x71 ;F
	 
	
	END                ; directive 'end of program'

